
/*
 *    miracleshaper - Traffic shaper and prioritizer for DSLs.
 *    Copyright (C) 2007  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

/* Here (and in pkt.c) you find all the functions for dissecting a packet,
 * or a struct pkt, to be exact.
 *
 * The basic procedure is that you call pkt_pinit(), which resets the
 * internal state of the struct pkt so that you are at the outermost
 * layer of the packet, that is, at the ethernet layer. Then you repeat
 * the following cycle for as many layers as you need:
 *
 * - Read any values you need from the headers of the current layer.
 * - In particular, find out what the next layer is.
 * - Then call the *_nextlayer function for the current layer, which
 *   will update the struct pkt so that the current layer is one layer
 *   deeper into the packet.
 *
 * Note that you can access fields of any layer only as long as it
 * is the current layer. If you need any values lateron, save them into
 * some temporary variables.
 *
 * In addition to the layer-specific functions, there is the generic
 * pkt_get_plen() function that returns the size of the packet at the
 * current layer. So, after every call to some *_nextlayer() function,
 * the value returned is reduced by the headers, footers, and padding
 * belonging to the outer layer you just left behind.
 *
 * All functions return some error indication if they encounter some
 * invalid value that prevents them from finding out what they are supposed
 * to find out--in particular if the packet is shorter than where the
 * field to read would be located. Functions returning pointers usually
 * return NULL in this case, functions returning integers usually
 * return -1.
 */

#ifndef PKT_H
#define PKT_H

#include <stdbool.h>
#include <netinet/in.h>

#include "config.h"
#include "utils.h"

#define PPP_P_IP 0x0021
#define PPP_P_IPV6 0x0057

struct pkt {
	ssize_t len,plen;
	char *pbuf;
#ifdef USEMARK
	uint32_t mark;
#endif
	char buf[0] __attribute__ ((aligned));
};

static inline ssize_t pkt_set_len(struct pkt *pkt,ssize_t len){
	return pkt->len=len;
}

static inline ssize_t pkt_get_len(struct pkt *pkt){
	return pkt->len;
}

#ifdef USEMARK
static inline void pkt_set_mark(struct pkt *pkt,uint32_t mark){
	pkt->mark=mark;
}
#endif

static inline char *pkt_get_buf(struct pkt *pkt){
	return pkt->buf;
}

static inline void pkt_pinit(struct pkt *pkt){
	pkt->plen=pkt->len;
	pkt->pbuf=pkt->buf;
}

static inline ssize_t pkt_set_plen(struct pkt *pkt,ssize_t len){
	return pkt->plen=len;
}

static inline ssize_t pkt_get_plen(struct pkt *pkt){
	return pkt->plen;
}

static inline void pkt_advance_pbuf(struct pkt *pkt,ssize_t x){
	pkt->pbuf+=x;
}

static inline char *pkt_get_pbuf(struct pkt *pkt){
	return pkt->pbuf;
}

/*** Ethernet ************************************************************/

static inline long long pkt_eth_get_dst(struct pkt *pkt){
	if(pkt_get_plen(pkt)<0+6)return -1;
	return mac_from_6oct((uint8_t *)pkt_get_pbuf(pkt)+0);
}

static inline long long pkt_eth_get_src(struct pkt *pkt){
	if(pkt_get_plen(pkt)<6+6)return -1;
	return mac_from_6oct((uint8_t *)pkt_get_pbuf(pkt)+6);
}

static inline long pkt_eth_get_proto(struct pkt *pkt){
	if(pkt_get_plen(pkt)<12+2)return -1;
	return ntohs(*(uint16_t *)(pkt_get_pbuf(pkt)+12));
}

static inline bool pkt_eth_nextlayer(struct pkt *pkt){
	if(pkt_set_plen(pkt,pkt_get_plen(pkt)-14)<0)return false;
	pkt_advance_pbuf(pkt,14);
	return true;
}

/*** PPPoE Session *******************************************************/

static inline bool pkt_pppoeses_nextlayer(struct pkt *pkt){
	if(pkt_set_plen(pkt,pkt_get_plen(pkt)-6)<0)return false;
	pkt_advance_pbuf(pkt,6);
	return true;
}

/*** PPP *****************************************************************/

static inline long pkt_ppp_get_proto(struct pkt *pkt){
	if(pkt_get_plen(pkt)<0+2)return -1;
	return ntohs(*(uint16_t *)(pkt_get_pbuf(pkt)+0));
}

static inline bool pkt_ppp_nextlayer(struct pkt *pkt){
	if(pkt_set_plen(pkt,pkt_get_plen(pkt)-2)<0)return false;
	pkt_advance_pbuf(pkt,2);
	return true;
}

/*** IP ******************************************************************/

bool pkt_ip_update_checksum(struct pkt *,unsigned int,unsigned int);

bool pkt_ip_mark_ecn(struct pkt *);

bool pkt_ip_mark_ecn_x(struct pkt *);

static inline int pkt_ip_get_headerlen(struct pkt *pkt){
	if(pkt_get_plen(pkt)<0+1)return -1;
	return ((*(uint8_t *)(pkt_get_pbuf(pkt)+0))&15)<<2;
}

static inline int pkt_ip_get_tos(struct pkt *pkt){
	if(pkt_get_plen(pkt)<1+1)return -1;
	return *(uint8_t *)(pkt_get_pbuf(pkt)+1)&0xfc;
}

static inline long pkt_ip_get_totallen(struct pkt *pkt){
	if(pkt_get_plen(pkt)<2+2)return -1;
	return ntohs(*(uint16_t *)(pkt_get_pbuf(pkt)+2));
}

static inline int pkt_ip_get_proto(struct pkt *pkt){
	if(pkt_get_plen(pkt)<9+1)return -1;
	return *(uint8_t *)(pkt_get_pbuf(pkt)+9);
}

static inline bool pkt_ip_nextlayer(struct pkt *pkt){
	ssize_t plen;
	int hlen;
	long totlen;
	
	plen=pkt_get_plen(pkt);
	hlen=pkt_ip_get_headerlen(pkt);
	totlen=pkt_ip_get_totallen(pkt);
	if(hlen<0||totlen<0||plen<totlen||totlen<hlen)return false;
	pkt_set_plen(pkt,totlen-hlen);
	pkt_advance_pbuf(pkt,hlen);
	return true;
}

/*** IPv6 ****************************************************************/

static inline long pkt_ip6_get_pllen(struct pkt *pkt){
	if(pkt_get_plen(pkt)<4+2)return -1;
	return ntohs(*(uint16_t *)(pkt_get_pbuf(pkt)+4));
}

static inline int pkt_ip6_get_proto(struct pkt *pkt){
	if(pkt_get_plen(pkt)<6+1)return -1;
	return *(uint8_t *)(pkt_get_pbuf(pkt)+6);
}

static inline int pkt_ip6_get_tc(struct pkt *pkt){
	if(pkt_get_plen(pkt)<2)return -1;
	return (ntohs(*(uint16_t *)pkt_get_pbuf(pkt))>>4)&0xfc;
}

static inline bool pkt_ip6_nextlayer(struct pkt *pkt){
	ssize_t plen;
	long pllen;

	plen=pkt_get_plen(pkt);
	pllen=pkt_ip6_get_pllen(pkt);
	if(pllen<0||plen<40+pllen)return false;
	pkt_set_plen(pkt,pllen);
	pkt_advance_pbuf(pkt,40);
	return true;
}

/*** UDP *****************************************************************/

static inline long pkt_udp_get_dstport(struct pkt *pkt){
	if(pkt_get_plen(pkt)<2+2)return -1;
	return ntohs(*(uint16_t *)(pkt_get_pbuf(pkt)+2));
}

/*** TCP *****************************************************************/

static inline int pkt_tcp_get_headerlen(struct pkt *pkt){
	if(pkt_get_plen(pkt)<12+1)return -1;
	return ((*(uint8_t *)(pkt_get_pbuf(pkt)+12))&0xf0)>>2;
}

static inline bool pkt_tcp_nextlayer(struct pkt *pkt){
	ssize_t plen;
	int hlen;

	plen=pkt_get_plen(pkt);
	hlen=pkt_tcp_get_headerlen(pkt);
	if(hlen<0||plen<hlen)return false;
	pkt_set_plen(pkt,plen-hlen);
	pkt_advance_pbuf(pkt,hlen);
	return true;
}

#endif

